/*
 * Copyright (c) 1991-2023 by STEP Tools Inc. 
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and
 * its documentation is hereby granted, provided that this copyright
 * notice and license appear on all copies of the software.
 * 
 * STEP TOOLS MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE
 * SUITABILITY OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING
 * BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT. STEP TOOLS
 * SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A
 * RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS
 * DERIVATIVES.
 *
 * Author: Blair Downie (downie@steptools.com)
 * Author: Dave Loffredo (loffredo@steptools.com)
 */

#include <stp_schema.h>
#include "assembly_utils.h"
#include "geometry_utils.h"
#include "unit_utils.h"

// Also provided by stixlib as stix_put_related_pdef()
void put_related_pdef(
    stp_product_definition_relationship * rel,
    stp_product_definition * pd
    )
{
    if (!rel) return;
    stp_product_definition_or_reference * sel = 
	rel-> related_product_definition();
    
    if (!sel) {
	sel = pnewIn(rel->design()) stp_product_definition_or_reference;
	rel-> related_product_definition(sel);
    }
    rose_put_nested_object (sel, pd);
}


// Also provided by stixlib as stix_put_relating_pdef()
void put_relating_pdef(
    stp_product_definition_relationship * rel,
    stp_product_definition * pd
    )
{
    if (!rel) return;
    stp_product_definition_or_reference * sel = 
	rel-> relating_product_definition();
    
    if (!sel) {
	sel = pnewIn(rel->design()) stp_product_definition_or_reference;
	rel-> relating_product_definition(sel);
    }
    rose_put_nested_object (sel, pd);
}

// Also provided by stixlib as stix_put_reprel_1()
void put_reprel_1(
    stp_representation_relationship * rel,
    stp_representation * rep
    )
{
    if (!rel) return;
    stp_representation_or_representation_reference * sel = rel-> rep_1();
    if (!sel) {
	sel = pnewIn(rel->design())
	    stp_representation_or_representation_reference;
	rel-> rep_1(sel);
    }
    rose_put_nested_object (sel, rep);
}

// Also provided by stixlib as stix_put_reprel_2()
void put_reprel_2(
    stp_representation_relationship * rel,
    stp_representation * rep
    )
{
    if (!rel) return;
    stp_representation_or_representation_reference * sel = rel-> rep_2();
    if (!sel) {
	sel = pnewIn(rel->design())
	    stp_representation_or_representation_reference;
	rel-> rep_2(sel);
    }
    rose_put_nested_object (sel, rep);
}

// Given two product_definitions this creates an assembly between them.
stp_next_assembly_usage_occurrence* createCMAsm(
    const char* asm_id,
    const char* asm_usage,
    const char* asm_desc,
    const char* ref_desig, 
    stp_product_definition* asm_pdef, 
    stp_product_definition* comp_pdef
    )
{
    // Create a next_assembly_usage occurrence to link the two.
    stp_next_assembly_usage_occurrence* nauo = 
	pnew stp_next_assembly_usage_occurrence();

    // The id has no standard mapping, but it should be unique.
    nauo->id (asm_id);

    // The name should contain something that identifies the usage.
    nauo->name (asm_usage);

    // There's no mapping for the description, "" is fine..
    nauo->description (asm_desc);

    // The reference_designator is optional.  When specified, it
    // should contain a unique location for the assembly.
    nauo->reference_designator (ref_desig);

    // The relating_product_definition is the assembly.
    put_relating_pdef (nauo, asm_pdef);

    // The related_product_definition is the component.
    put_related_pdef(nauo, comp_pdef);

    return nauo;
}

// This approach can be used if both of the shape descriptions are the
// same type (same shape rep subtype). Given two shape_representations
// and their axis2_placements, this creates the geometric portion of
// the assembly using a mapped_item.
//
void createGeomAsmMI(
    stp_representation* asmSR, 
    stp_representation_item* asmAxis,
    stp_representation* compSR,
    stp_representation_item* compAxis,
    stp_next_assembly_usage_occurrence* nauo
    )
{
    /* We pass in representations to avoid having to do casts,
     * but this should only be used to relate instances of the
     * shape representation subtype.
     */
    if (!asmSR->  isa (ROSE_DOMAIN(stp_shape_representation)) ||
	!compSR-> isa (ROSE_DOMAIN(stp_shape_representation)))
    {
	printf ("createGeomAsmMI: representations must be instances of");
	printf ("  the shape representation subtype");
	return;
    }


    // Create a mapped item & representation map for the component.
    stp_mapped_item* mi = pnew stp_mapped_item();
    stp_representation_map* rm = pnew stp_representation_map();

    // The name of the mapped item has no standard mapping.
    mi->name("");

    // The source of the mapped_item is the representation_map.
    mi->mapping_source(rm);

    // The target is the assembly's axis.
    mi->mapping_target(asmAxis);

    // The mapping origin is the components coordinate system.
    rm->mapping_origin(compAxis);

    // The mapped_representation is the component shape_representation.
    rm->mapped_representation(compSR);

    // Now a new shape_representation is needed for the
    // next_assembly_usage_occurrence.  This shape_representation will
    // reference the mapped_item used in this assembly so that one can
    // navigate from the nauo to the mapped_item.  This is important
    // because a component can be referenced more than once in an
    // assembly (left tire, right tire) and this way the
    // correspondence between the nauo and mapped_item for each
    // instance can be maintained.  Create the same kind of shape_rep
    // as the others.
    stp_shape_representation* nauoSR = 
	ROSE_CAST(stp_shape_representation, asmSR->domain()->pnewInstance());

    // There's no mapping for the representation name.
    nauoSR->name("");

    // We'll reuse the context from the assembly.
    nauoSR->context_of_items(asmSR->context_of_items());

    // Create a new items list.
    SetOfstp_representation_item* items = pnew SetOfstp_representation_item();
    nauoSR->items(items);

    // Reference the mapped_item.
    items->add(mi);

    // now attach the shape_representation to the nauo using a
    // product_definition_shape and a shape_definition_representation.
    stp_product_definition_shape* pds = pnew stp_product_definition_shape();
    stp_shape_definition_representation* sdr = 
	pnew stp_shape_definition_representation();

    // There's no standard mapping for the name or description.
    pds->name("");
    pds->description("");

    // The definition should point to next_assembly_usage_occurrence.
    stp_characterized_definition* cd = pnew stp_characterized_definition();
    stp_characterized_product_definition* cpd = 
	pnew stp_characterized_product_definition();
    cd->_characterized_product_definition(cpd);
    cpd->_product_definition_relationship(nauo);
    pds->definition(cd);

    // The definition of the shape_definition_representation is the
    // product_definition_shape.  This reference goes through a select
    // type called represented_definition.  
    //
    sdr-> definition(pnew stp_represented_definition);
    sdr-> definition()-> _property_definition(pds);

    // The representation is the new one we've created.
    sdr->used_representation(nauoSR);
}

// This method should be used if the shape types are different.  Given
// two shape_representations and their axis2_placements, this creates
// the geometric portion of the assembly using a
// context_dependent_shape_representation.
//
void createGeomAsmCDSR(
    stp_representation* asmSR, 
    stp_representation_item* asmAxis,
    stp_representation* compSR,
    stp_representation_item* compAxis,
    stp_next_assembly_usage_occurrence* nauo
    )
{
    /* We pass in representations to avoid having to do casts,
     * but this should only be used to relate instances of the
     * shape representation subtype.
     */
    if (!asmSR->  isa (ROSE_DOMAIN(stp_shape_representation)) ||
	!compSR-> isa (ROSE_DOMAIN(stp_shape_representation)))
    {
	printf ("createGeomAsmCDSR: representations must be instances of");
	printf ("  the shape representation subtype");
	return;
    }

    // Create a product_definition_shape to link the cdsr to the nauo
    stp_product_definition_shape* pds = pnew stp_product_definition_shape();

    // There's no standard mapping for the name or description.
    pds->name("");
    pds->description("");

    // The definition should point to next_assembly_usage_occurrence.
    stp_characterized_definition* cd = pnew stp_characterized_definition();
    stp_characterized_product_definition* cpd = 
	pnew stp_characterized_product_definition();
    cd->_characterized_product_definition(cpd);
    cpd->_product_definition_relationship(nauo);
    pds->definition(cd);

    // Create a context_dependent_shape_representation.
    stp_context_dependent_shape_representation* cdsr = 
	pnew stp_context_dependent_shape_representation();

    // The represented_product_relation is the pds.
    cdsr->represented_product_relation(pds);

    // A complex entity is used for the shape_representation_relationship.
    stp_representation_relationship_with_transformation_and_shape_representation_relationship * repRel = 
	pnew stp_representation_relationship_with_transformation_and_shape_representation_relationship();
    cdsr->representation_relation(repRel);

    // The name and description attributes have no standard mapping
    repRel->name("");
    repRel->description("");

    // rep_1 is the assembly shape and rep_2 is the component.
    put_reprel_1(repRel, asmSR);
    put_reprel_2(repRel, compSR);

    // The transformation_operator should be an item_defined_transform.
    stp_item_defined_transformation* xform = pnew stp_item_defined_transformation();
    stp_transformation* trans = pnew stp_transformation();
    trans->_item_defined_transformation(xform);
    repRel->transformation_operator(trans);

    // The name and description fields have no standard mapping.
    xform->name("");
    xform->description("");

    // The first transform_item is the axis of the assembly and the second is
    // the component.
    xform->transform_item_1(asmAxis);
    xform->transform_item_2(compAxis);
}


stp_product_definition * get_property_pdef (
    stp_property_definition_representation * pdr
    )
{
    /* Find the product_definition that a property is attached to.
     * Note that properties can be attached to other things besides a
     * product definition.  This also works on property definition
     * representation subtypes like shape_definition_representation.
     */

    /* The pd_representation has a definition that usually refers to a
     * property_definition. 
     */
    if (!pdr->definition()) return 0;
    if (!pdr->definition()-> is_property_definition()) 
	return 0;

    /* The property_definition has a definition that usually refers to
     * a product_definition, but could refer to other things.  The link
     * to product_definition goes through another select type called
     * characterized_product_definition, so we have to look down two
     * levels.  Refer to the HTML version of the EXPRESS schema for a
     * better picture of how these are nested.
     */
    stp_property_definition * prop = 
	pdr->definition()-> _property_definition();

    stp_characterized_definition * cd = prop-> definition();
    if (cd-> is_characterized_product_definition()) 
    {
	stp_characterized_product_definition * cpd =
	    cd-> _characterized_product_definition();

	if (cpd-> is_product_definition()) 
	    return cpd-> _product_definition();
    }
    return 0;
}
